#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/kthread.h>

int count;
DEFINE_SEQLOCK(count_lock);
struct task_struct *read_thread1, *read_thread2, *read_thread3, *write_thread;

#define SLEEP_MS 500

static int write_func(void *data)
{
  while (!kthread_should_stop()) {
    write_seqlock(&count_lock);
    count++;
    write_sequnlock(&count_lock);
    msleep(SLEEP_MS);
  }

  do_exit(0);
}

static int read_func(void *data)
{
  unsigned long seq;

  while (!kthread_should_stop()) {
    do {
      seq = read_seqbegin(&count_lock);
    } while (read_seqretry(&count_lock, seq));

    printk("count = %d\n", count);
    msleep(SLEEP_MS);
  }

  do_exit(0);
}

static int __init seqlock_kernel_module_init(void)
{
  printk("seqlock_kernel_module_init()\n");
  count = 0;

  read_thread1 = kthread_run(read_func, NULL, "read-thread1");
  read_thread2 = kthread_run(read_func, NULL, "read-thread2");
  read_thread3 = kthread_run(read_func, NULL, "read-thread3");
  write_thread = kthread_run(write_func, NULL, "write-thread");
  return 0;
}

static void __exit seqlock_kernel_module_exit(void)
{
  printk("seqlock_kernel_module_exit()\n");
  kthread_stop(read_thread3);
  kthread_stop(read_thread2);
  kthread_stop(read_thread1);
  kthread_stop(write_thread);
}

module_init(seqlock_kernel_module_init);
module_exit(seqlock_kernel_module_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Hiroyuki Chishiro");
MODULE_DESCRIPTION("Seqlock Kernel Module");
